using UnityEngine;
using UnityEngine.Events;
using System;
using System.Collections.Generic;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using Battlehub.RTSaveLoad;

namespace Ruvik
{
    class PH : MVRScript
    {
        string ASSETBUNDLEPATH;
        static Keybind activationKb, switchMotionKb, switchAxisKb;
        Dictionary<string, Keybind> keyBindPairs = new Dictionary<string, Keybind>();
        static bool isRecording;
        string keyString = string.Empty;
        public static JSONStorableFloat sensetivity, rotationSpeed,UIScale,UIOffsetX,UIOffsetY;
        public static JSONStorableBool adjustPosition, alignUpDown,changePose;

        public void OnBindingsListRequested(List<object> bindings)
        {
            bindings.Add(new JSONStorableAction("activate/deactive UI", () => PHUI.menuUIActive = !PHUI.menuUIActive));
            bindings.Add(new JSONStorableAction("clear selected bones", () => PHUI.ClearActiveNodes()));
            bindings.Add(new JSONStorableAction("ik link", () => PHUtils.IKLink()));
            bindings.Add(new JSONStorableAction("reset linking state", () => PHUtils.ResetLink_method()));
            bindings.Add(new JSONStorableAction("reset pose", () => PHUtils.ResetPose_method()));
            bindings.Add(new JSONStorableAction("undo", () => PHUndo.RestoreState()));
            bindings.Add(new JSONStorableAction("toggle help", () => {PHUI.phHelpMessageObject.SetActive(!PHUI.phHelpMessageObject.activeInHierarchy); }));
        }
        public static string GetPluginPath(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }
        void Update()
        {
            ProccesKeyBindings();
            CheckForKeybindUpdate();
            if (PHUI.menuUIActive)
                PHUI.PHMenuSetActive(SuperController.singleton.MainHUDVisible);
            else
                PHUI.PHMenuSetActive(false);
        }
        public override void Init()
        {
            sensetivity = new JSONStorableFloat("sensetivity", 1.0f, 0.1f, 4.0f);
            rotationSpeed = new JSONStorableFloat("rotation Speed", 2.5f, 0.1f, 10.0f);
            UIScale = new JSONStorableFloat("UI Scale", 1.0f, 0.5f, 2.0f);
            UIOffsetX = new JSONStorableFloat("UI Offset X", 10.0f, 0.0f, 50.0f);
            UIOffsetY = new JSONStorableFloat("UI Offset Y", 10.0f, 0.0f, 50.0f);
            sensetivity.storeType = JSONStorableParam.StoreType.Full;
            rotationSpeed.storeType = JSONStorableParam.StoreType.Full;
            UIScale.storeType = JSONStorableParam.StoreType.Full;
            UIOffsetX.storeType = JSONStorableParam.StoreType.Full;
            UIOffsetY.storeType = JSONStorableParam.StoreType.Full;

            SuperController.singleton.BroadcastMessage("OnActionsProviderAvailable", this, SendMessageOptions.DontRequireReceiver);
            ASSETBUNDLEPATH = GetPluginPath(this) + "/assets/PH.assetbundle";
            Request request = new Request { path = ASSETBUNDLEPATH, callback = OnAssetBundleLoaded };
            AssetLoader.QueueLoadAssetBundleFromFile(request);

            GameObject identity = new GameObject();
            identity.transform.position = Vector3.zero;
            identity.transform.rotation = Quaternion.identity;
            identity.transform.localScale = Vector3.one;
            PHUtils.motionTransform = new Transform[3] { null, identity.transform, SuperController.singleton.lookCamera.transform };

            PHLinker.InitGraph();

            RegisterAction(new JSONStorableAction("activate/deactive UI", () => PHUI.menuUIActive = !PHUI.menuUIActive));
            RegisterAction(new JSONStorableAction("clear selected bones", () => PHUI.ClearActiveNodes()));
            RegisterAction(new JSONStorableAction("ik link", () => PHUtils.IKLink()));
            RegisterAction(new JSONStorableAction("reset linking state", () => PHUtils.ResetLink_method()));
            RegisterAction(new JSONStorableAction("reset pose", () => PHUtils.ResetPose_method()));
            RegisterAction(new JSONStorableAction("undo", () => PHUndo.RestoreState()));

            var btn = CreateButton("Show UI", false); btn.button.onClick.AddListener(delegate () { PHUI.menuUIActive = true; });
            CreateSpacer(true).height = btn.height;


            activationKb = GenKey("start the motion", KeyCode.Space);
            switchMotionKb = GenKey("switch the motion(position to rotation)", KeyCode.LeftShift);
            switchAxisKb = GenKey("switch the axis", KeyCode.LeftControl);

            CreateSpacer(true).height = 20;
            CreateSpacer(false).height = 20;


            CreateSlider(sensetivity, false);
            CreateSlider(rotationSpeed, true);

            RegisterFloat(sensetivity);
            RegisterFloat(rotationSpeed);

            CreateSpacer(true).height = 20;
            CreateSpacer(false).height = 20;

            PHUtils.xPosMultiplier = CreateInverseButton("invert X Pos", false, delegate (bool a) { PHUtils.xPosMultiplier = a ? -1.0f : 1.0f; });
            PHUtils.yPosMultiplier = CreateInverseButton("invert Y Pos", false, delegate (bool a) { PHUtils.yPosMultiplier = a ? -1.0f : 1.0f; });
            PHUtils.zPosMultiplier = CreateInverseButton("invert Z Pos", false, delegate (bool a) { PHUtils.zPosMultiplier = a ? -1.0f : 1.0f; });
            PHUtils.xRotMultiplier = CreateInverseButton("invert X Rot", true, delegate (bool a) { PHUtils.xRotMultiplier = a ? -1.0f : 1.0f; });
            PHUtils.yRotMultiplier = CreateInverseButton("invert Y Rot", true, delegate (bool a) { PHUtils.yRotMultiplier = a ? -1.0f : 1.0f; });
            PHUtils.zRotMultiplier = CreateInverseButton("invert Z Rot", true, delegate (bool a) { PHUtils.zRotMultiplier = a ? -1.0f : 1.0f; });

            CreateSpacer(true).height = 20;
            CreateSpacer(false).height = 20;

            adjustPosition = new JSONStorableBool("adjust_position", true);
            adjustPosition.storeType = JSONStorableParam.StoreType.Full;
            RegisterBool(adjustPosition);
            CreateToggle(adjustPosition, true);

            alignUpDown = new JSONStorableBool("align_up_down", true);
            alignUpDown.storeType = JSONStorableParam.StoreType.Full;
            RegisterBool(alignUpDown);
            CreateToggle(alignUpDown, false);


            CreateSpacer(true).height = 20;
            CreateSpacer(false).height = 20;


            CreateSlider(UIScale, false).slider.onValueChanged.AddListener((float val) => {GameObject.Find("PH_root").GetComponent<RectTransform>().localScale = Vector3.one * val; });
            CreateSlider(UIOffsetX, true).slider.onValueChanged.AddListener((float val) => { GameObject.Find("PH_root").GetComponent<RectTransform>().anchoredPosition = new Vector3(-val,UIOffsetY.valNoCallback,  0); } );
            CreateSlider(UIOffsetY, true).slider.onValueChanged.AddListener((float val) => { GameObject.Find("PH_root").GetComponent<RectTransform>().anchoredPosition = new Vector3(-UIOffsetX.valNoCallback,val,  0); } );
            changePose = new JSONStorableBool("change Pose", true);
            changePose.storeType = JSONStorableParam.StoreType.Full;
            RegisterBool(changePose);
            CreateToggle(changePose, false);

            RegisterFloat(UIScale);
            RegisterFloat(UIOffsetX);
            RegisterFloat(UIOffsetY);
        }
        public float CreateInverseButton(string name, bool right, UnityAction<bool> toggleAction)
        {
            JSONStorableBool storableBool = new JSONStorableBool(name, false);
            storableBool.storeType = JSONStorableParam.StoreType.Full;
            RegisterBool(storableBool);
            CreateToggle(storableBool, right);
            storableBool.toggle.onValueChanged.AddListener(toggleAction);
            return storableBool.val ? 1.0f : -1.0f;
        }
        public Keybind GenKey(string actionName, KeyCode defaultKey)
        {
            CreateButton(actionName, false);
            var btn = CreateButton(defaultKey.ToString(), true);
            Keybind kb = new Keybind(defaultKey, btn);
            keyBindPairs.Add(actionName, kb);
            btn.button.onClick.AddListener(delegate () { keyString = actionName; btn.label = "...."; btn.buttonColor = Color.red; isRecording = true; });
            return kb;
        }
        public void FetchKeyOrButton(Keybind kb)
        {
            foreach (KeyCode kcode in Enum.GetValues(typeof(KeyCode)))
            {
                if (Input.GetKeyDown(kcode))
                {
                    kb.Update(kcode);
                    break;
                }
            }
        }
        public class Keybind
        {
            public KeyCode key;
            UIDynamicButton UIbtn;
            public Keybind(KeyCode defaultKey, UIDynamicButton btn)
            {
                key = defaultKey;
                UIbtn = btn;
            }
            public void Update(KeyCode key)
            {
                string keyAsString = key.ToString();
                isRecording = false;
                UIbtn.buttonColor = new Color(0.78f, 0.78f, 0.78f);
                this.UIbtn.label = keyAsString;
                this.key = key;
            }
        }
        void ProccesKeyBindings()
        {

            if (Input.GetKeyDown(activationKb.key))
                PHUtils.PreMoveControllers();

            if (Input.GetKey(activationKb.key))
            {
                float x, y;
                Vector3 moveby;
                x = Input.GetAxis("Mouse X");
                y = Input.GetAxis("Mouse Y");
                moveby = Input.GetKey(switchAxisKb.key) ? new Vector3(0, 0, x + y) : new Vector3(x, y, 0);
                PHUtils.MoveActiveNodes(moveby,!Input.GetKey(switchMotionKb.key));
            }
            if (Input.GetKeyUp(activationKb.key))
                PHUtils.PostMoveControllers();

        }
        private void CheckForKeybindUpdate()
        {
            if (isRecording)
                FetchKeyOrButton(keyBindPairs.Get(keyString));

        }
        public void OnDestroy()
        {

            AssetLoader.DoneWithAssetBundleFromFile(ASSETBUNDLEPATH);
            SuperController.singleton.BroadcastMessage("OnActionsProviderDestroyed", this, SendMessageOptions.DontRequireReceiver);
            PHUI.PHMenuDesroy();
        }
        public void OnDisable()
        {
            PHUI.PHMenuSetActive(false);
        }
        public void OnEnable()
        {
            PHUI.PHMenuSetActive(true);
        }
        private void OnAssetBundleLoaded(Request aRequest)
        {
            var PHUIAsset = aRequest.assetBundle.LoadAsset<GameObject>("Assets/ph/posinghelperui.prefab");
            var off_off = aRequest.assetBundle.LoadAsset<Sprite>("Assets/ph/off_off.png");
            var off_on = aRequest.assetBundle.LoadAsset<Sprite>("Assets/ph/off_on.png");
            var on_off = aRequest.assetBundle.LoadAsset<Sprite>("Assets/ph/on_off.png");
            var on_on = aRequest.assetBundle.LoadAsset<Sprite>("Assets/ph/on_on.png");
            if (PHUIAsset != null)
            {
                PHUI.phGameObject = Instantiate(PHUIAsset);
                PHUI.stateTextures.Add("off_off",off_off);
                PHUI.stateTextures.Add("off_on",off_on);
                PHUI.stateTextures.Add("on_off",on_off);
                PHUI.stateTextures.Add("on_on",on_on);
                PHUI.InitPHUI();
            }
        }

    }
}
