﻿using Battlehub.RTSaveLoad;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Ruvik
{
    class PHLinker
    {
        public static List<string> graphDic = new List<string>
        {
            "headControl",
            "neckControl",
            "chestControl",
            "rArmControl",
            "rElbowControl",
            "rHandControl",
            "lArmControl",
            "lElbowControl",
            "lHandControl",
            "abdomen2Control",
            "pelvisControl",
            "hipControl",
            "abdomenControl",
            "penisBaseControl",
            "penisMidControl",
            "penisTipControl",
            "rThighControl",
            "rKneeControl",
            "rFootControl",
            "rToeControl",
            "lThighControl",
            "lKneeControl",
            "lFootControl",
            "lToeControl",
        };
        static int VERTEX_COUNT = 24;
        static List<List<int>> graph = new List<List<int>>(VERTEX_COUNT);
        public static int GetPoint(string p)
        {
            return graphDic.IndexOf(p);
        }
        public static void InitGraph()
        {
            for (int i = 0; i < VERTEX_COUNT; i++)
            {
                graph.Add(new List<int>());
            }
            addEdge("head", "neck");
            addEdge("neck", "chest");
            addEdge("chest", "abdomen2");
            addEdge("abdomen2", "pelvis");
            addEdge("hip", "pelvis");
            addEdge("hip", "abdomen");
            addEdge("abdomen", "penisBase");
            addEdge("penisBase", "penisMid");
            addEdge("penisMid", "penisTip");
            addEdge("chest", "lArm");
            addEdge("lArm", "lElbow");
            addEdge("lElbow", "lHand");
            addEdge("chest", "rArm");
            addEdge("rArm", "rElbow");
            addEdge("rElbow", "rHand");
            addEdge("abdomen", "lThigh");
            addEdge("lThigh", "lKnee");
            addEdge("lKnee", "lFoot");
            addEdge("lFoot", "lToe");
            addEdge("abdomen", "rThigh");
            addEdge("rThigh", "rKnee");
            addEdge("rKnee", "rFoot");
            addEdge("rFoot", "rToe");
        }
        public static void LinkPath(Node[] path)
        {
            for (int i = path.Length - 1;
                    i > 0; i--)
            {
                FreeControllerV3 child = path[i].controller;
                FreeControllerV3 parent = path[i-1].controller;
                PHUtils.Link(child, parent);
                if (Input.GetKey(KeyCode.LeftShift))
                {
                    if (child.name[0] == 'r' || child.name[0] == 'l')
                    {
                        child.currentPositionState = FreeControllerV3.PositionState.Off;
                    }
                }
            }

        }
        public static Node[] FindPath(FreeControllerV3 a, FreeControllerV3 b)
        {
            int source = GetPoint(a.name), dest = GetPoint(b.name);
            return FindPath(source, dest);
        }

        private static void addEdge(string a, string b)
        {
            int i = GetPoint(a + "Control");
            int j = GetPoint(b + "Control");
            graph[i].Add(j);
            graph[j].Add(i);
        }

        private static Node[] FindPath(int s, int dest)
        {
            int[] pred = new int[VERTEX_COUNT];
            int[] dist = new int[VERTEX_COUNT];

            BFS(s, dest, pred, dist);

            List<int> path = new List<int>();
            int crawl = dest;
            path.Add(crawl);

            while (pred[crawl] != -1)
            {
                path.Add(pred[crawl]);
                crawl = pred[crawl];
            }


            Node[] nodes = new Node[path.Count];
            for (int i = 0; i < path.Count; i++)
            {
                nodes[i] = PHUI.GetNodeByName(graphDic[path[i]].Split('C')[0]);
            }
            return nodes;
        }

        private static bool BFS(int src, int dest, int[] pred, int[] dist)
        {
            List<int> queue = new List<int>();

            bool[] visited = new bool[VERTEX_COUNT];

            for (int i = 0; i < VERTEX_COUNT; i++)
            {
                visited[i] = false;
                dist[i] = int.MaxValue;
                pred[i] = -1;
            }

            visited[src] = true;
            dist[src] = 0;
            queue.Add(src);

            while (queue.Count != 0)
            {
                int u = queue[0];
                queue.RemoveAt(0);

                for (int i = 0;
                        i < graph[u].Count; i++)
                {
                    if (visited[graph[u][i]] == false)
                    {
                        visited[graph[u][i]] = true;
                        dist[graph[u][i]] = dist[u] + 1;
                        pred[graph[u][i]] = u;
                        queue.Add(graph[u][i]);

                        if (graph[u][i] == dest)
                            return true;
                    }
                }
            }
            return false;
        }
    }
}