using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Ruvik
{
    class PHUndo
    {
        class BonesState
        {
            public Vector3 position;
            public Vector3 rotation;
            public FreeControllerV3.PositionState pstate;
            public FreeControllerV3.RotationState rstate;
        }
        static List<BonesState[]> states = new List<BonesState[]>();
        public static void ClearCache()
        {
            states.Clear();
        }
        static private void RestoreBone(ref FreeControllerV3 bone , BonesState boneState)
        {
            bone.currentPositionState = FreeControllerV3.PositionState.On;
            bone.currentRotationState = FreeControllerV3.RotationState.On;
            bone.RotateControl(boneState.rotation);
            bone.MoveControl(boneState.position);
            bone.currentPositionState = boneState.pstate;
            bone.currentRotationState = boneState.rstate;
        }
        static private BonesState StoreBone(FreeControllerV3 bone)
        {
            BonesState bs = new BonesState();
            bs.position = bone.transform.position;
            bs.rotation = bone.transform.eulerAngles;
            bs.pstate = bone.currentPositionState;
            bs.rstate = bone.currentRotationState;
            return bs;
        }
        static public void RestoreState()
        {
            if (states.Count <= 0)
                return;
            int storeIndex = 0;
            BonesState[] state = states.Last();
            RestoreBone(ref PHUtils.activeAtom.freeControllers[0], state[storeIndex]);
            storeIndex++;
            RestoreBone(ref PHUtils.head, state[storeIndex]);
            storeIndex++;
            RestoreBone(ref PHUtils.neck, state[storeIndex]);
            storeIndex++;
            for (int i = 0; i < 2; i++)
            {
                RestoreBone(ref PHUtils.shoulders[i], state[storeIndex]);
                storeIndex++;
                RestoreBone(ref PHUtils.elbows[i], state[storeIndex]);
                storeIndex++;
                RestoreBone(ref PHUtils.hands[i], state[storeIndex]);
                storeIndex++;
            }
            RestoreBone(ref PHUtils.chest, state[storeIndex]);
            storeIndex++;
            RestoreBone(ref PHUtils.abdomen2, state[storeIndex]);
            storeIndex++;
            RestoreBone(ref PHUtils.hip, state[storeIndex]);
            storeIndex++;
            RestoreBone(ref PHUtils.abdomen1, state[storeIndex]);
            storeIndex++;
            for (int i = 0; i < 2; i++)
            {
                RestoreBone(ref PHUtils.thighs[i], state[storeIndex]);
                storeIndex++;
                RestoreBone(ref PHUtils.knees[i], state[storeIndex]);
                storeIndex++;
                RestoreBone(ref PHUtils.feet[i], state[storeIndex]);
                storeIndex++;
            }
            states.Remove(states.Last());
        }
        static public void StoreState()
        {
            int storeIndex = 0;
            BonesState[] state = new BonesState[19];
            state[storeIndex] = StoreBone(PHUtils.activeAtom.freeControllers[0]);
            storeIndex++;
            state[storeIndex] = StoreBone(PHUtils.head);
            storeIndex++;
            state[storeIndex] = StoreBone(PHUtils.neck);
            storeIndex++;
            for (int i = 0; i < 2; i++)
            {
                state[storeIndex] = StoreBone(PHUtils.shoulders[i]);
                storeIndex++;
                state[storeIndex] = StoreBone(PHUtils.elbows[i]);
                storeIndex++;
                state[storeIndex] = StoreBone(PHUtils.hands[i]);
                storeIndex++;
            }
            state[storeIndex] = StoreBone(PHUtils.chest);
            storeIndex++;
            state[storeIndex] = StoreBone(PHUtils.abdomen2);
            storeIndex++;
            state[storeIndex] = StoreBone(PHUtils.hip);
            storeIndex++;
            state[storeIndex] = StoreBone(PHUtils.abdomen1);
            storeIndex++;
            for (int i = 0; i < 2; i++)
            {
                state[storeIndex] = StoreBone(PHUtils.thighs[i]);
                storeIndex++;
                state[storeIndex] = StoreBone(PHUtils.knees[i]);
                storeIndex++;
                state[storeIndex] = StoreBone(PHUtils.feet[i]);
                storeIndex++;
            }
            if (states.Count >= 100)
                states.Remove(states.First());
            states.Add(state);
        }
    }
}
