﻿using MVR.FileManagementSecure;
using SimpleJSON;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.Rendering;
using static MeshVR.PresetManager;

namespace Ruvik
{
    class PHUtils
    {
        static public FreeControllerV3 head;
        static public FreeControllerV3 neck;
        static public FreeControllerV3[] shoulders;
        static public FreeControllerV3[] elbows;
        static public FreeControllerV3[] hands;
        static public FreeControllerV3 chest;
        static public FreeControllerV3 abdomen2;
        static public FreeControllerV3 pelvis;
        static public FreeControllerV3 hip;
        static public FreeControllerV3 abdomen1;
        static public FreeControllerV3[] thighs;
        static public FreeControllerV3[] knees;
        static public FreeControllerV3[] feet;
        static public FreeControllerV3[] toes;
        static public FreeControllerV3[] penis;
        public static List<Node> activeNodes = new List<Node>();
        public static Atom activeAtom = null;
        static FreeControllerV3.PositionState[] currentPStates;
        static FreeControllerV3.RotationState[] currentRStates;
        public static Transform[] motionTransform;
        public static int motionTransformIndex = 0;

        public static float xPosMultiplier = 1.0f;
        public static float yPosMultiplier = 1.0f;
        public static float zPosMultiplier = 1.0f;
        public static float xRotMultiplier = 1.0f;
        public static float yRotMultiplier = 1.0f;
        public static float zRotMultiplier = 1.0f;

        static public void ToggleSymmetry()
        {
            PHUI.isSymmetryToggled = !PHUI.isSymmetryToggled;
            PHUI.symmetryImage.color = PHUI.isSymmetryToggled ? Color.green:Color.white;
            if(PHUI.isSymmetryToggled)
                PHUI.ActivateRightSide(false);
            else
                PHUI.ActivateRightSide(true);
        }
        static public void ToggleState()
        {
            activeAtom.ToggleOn();
            PHUI.onOffImage.color = activeAtom.on ? Color.green : Color.white;
        }
        static public void ToggleFreeze()
        {
            activeAtom.SetFreezePhysics(!activeAtom.isPhysicsFrozen);
            PHUI.freezeImage.color = activeAtom.isPhysicsFrozen ? Color.green : Color.white;
        }
        static public void ToggleCollision()
        {
            activeAtom.SetCollisionEnabled(!activeAtom.collisionEnabled);
            PHUI.collisionImage.color = activeAtom.collisionEnabled ? Color.green : Color.white;
        }

        static string posePath = "Custom/Atom/Person/Pose";
        static List<ShortCut> poseShortcuts = FileManagerSecure.GetShortCutsForDirectory(posePath, false, false, true, true);
        static public void LoadPose()
        {
            PHUndo.StoreState();
            SuperController.singleton.GetMediaPathDialog(HandleLoadPreset,"vap", posePath, false, true, false, null, false, poseShortcuts);
            //activeAtom.LoadPhysicalPresetDialog();
        }
        static public void ResetPose_method()
        {
            PHUndo.StoreState();
            for (int i = 1; i < activeAtom.freeControllers.Length; i++)
            {
                activeAtom.freeControllers[i].Reset();
            }
            var morphControl = ((activeAtom.GetStorableByID("geometry")) as DAZCharacterSelector).morphsControlUI;

            morphControl.GetMorphUids().ForEach((uid) =>
            {
                var morph = morphControl.GetMorphByUid(uid);
                if (morph.isPoseControl)
                {
                    morph.SetDefaultValue();
                }
            });
        }
        static public void ChangeAlignment()
        {
            motionTransformIndex++;
            motionTransformIndex = motionTransformIndex % 3;
        }
        public static void UpdateNodes(Atom atom)
        {
            if (activeNodes == null)
                activeNodes = new List<Node>();
            PHUI.ClearActiveNodes();
            activeAtom = atom;
            motionTransform[0] = activeAtom.freeControllers[0].transform;
            knees = atom.freeControllers.Where(fc => fc.name.Contains("KneeControl")).ToArray();
            hands = atom.freeControllers.Where(fc => fc.name.Contains("HandControl")).ToArray();
            feet = atom.freeControllers.Where(fc => fc.name.Contains("FootControl")).ToArray();
            toes = atom.freeControllers.Where(fc => fc.name.Contains("ToeControl")).ToArray();
            thighs = atom.freeControllers.Where(fc => fc.name.Contains("ThighControl")).ToArray();
            elbows = atom.freeControllers.Where(fc => fc.name.Contains("ElbowControl")).ToArray();
            shoulders = atom.freeControllers.Where(fc => fc.name.Contains("ArmControl")).ToArray();
            penis = atom.freeControllers.Where(fc => fc.name.Contains("penis")).ToArray();
            hip = atom.freeControllers.First(fc => fc.name.Contains("hip"));
            chest = atom.freeControllers.First(fc => fc.name.Contains("chest"));
            head = atom.freeControllers.First(fc => fc.name.Contains("head"));
            neck = atom.freeControllers.First(fc => fc.name.Contains("neck"));
            abdomen1 = atom.freeControllers.First(fc => fc.name == "abdomenControl");
            abdomen2 = atom.freeControllers.First(fc => fc.name == "abdomen2Control");
            pelvis = atom.freeControllers.First(fc => fc.name == "pelvisControl");

            PHUI.collisionImage.color = activeAtom.collisionEnabled ? Color.green : Color.white;
            PHUI.freezeImage.color = activeAtom.isPhysicsFrozen ? Color.green : Color.white;
            PHUI.onOffImage.color = activeAtom.on ? Color.green : Color.white;

            PHUI.UpdateNode("node_head", head);
            PHUI.UpdateNode("node_neck", neck);
            PHUI.UpdateNode("node_chest", chest);
            PHUI.UpdateNode("node_abdomen2", abdomen2);
            PHUI.UpdateNode("node_hip", hip);
            PHUI.UpdateNode("node_pelvis", pelvis);
            PHUI.UpdateNode("node_abdomen", abdomen1);
            PHUI.UpdateNode("node_penisBase", penis[0]);
            PHUI.UpdateNode("node_penisMid", penis[1]);
            PHUI.UpdateNode("node_penisTip", penis[2]);
            PHUI.UpdateNode("select_name", activeAtom.freeControllers[0]);
            PHUI.UpdateNodeAlt("node_rToe", toes[0], "node_lToe", toes[1]);
            PHUI.UpdateNodeAlt("node_rFoot", feet[0], "node_lFoot", feet[1]);
            PHUI.UpdateNodeAlt("node_rKnee", knees[0], "node_lKnee", knees[1]);
            PHUI.UpdateNodeAlt("node_rThigh", thighs[0], "node_lThigh", thighs[1]);
            PHUI.UpdateNodeAlt("node_rArm", shoulders[0], "node_lArm", shoulders[1]);
            PHUI.UpdateNodeAlt("node_rElbow", elbows[0], "node_lElbow", elbows[1]);
            PHUI.UpdateNodeAlt("node_rHand", hands[0], "node_lHand", hands[1]);
        }

        public static void ResetLink_method()
        {
            PHUndo.StoreState();
            foreach (FreeControllerV3 fc in activeAtom.freeControllers)
            {
                fc.SelectLinkToRigidbody(null);
            }
            head.currentPositionState = FreeControllerV3.PositionState.On;
            head.currentRotationState = FreeControllerV3.RotationState.On;

            neck.currentPositionState = FreeControllerV3.PositionState.Off;
            neck.currentRotationState = FreeControllerV3.RotationState.Off;

            shoulders[0].currentPositionState = FreeControllerV3.PositionState.Off;
            shoulders[1].currentPositionState = FreeControllerV3.PositionState.Off;
            shoulders[0].currentRotationState = FreeControllerV3.RotationState.Off;
            shoulders[1].currentRotationState = FreeControllerV3.RotationState.Off;

            elbows[0].currentPositionState = FreeControllerV3.PositionState.Off;
            elbows[1].currentPositionState = FreeControllerV3.PositionState.Off;
            elbows[0].currentRotationState = FreeControllerV3.RotationState.Off;
            elbows[1].currentRotationState = FreeControllerV3.RotationState.Off;

            hands[0].currentPositionState = FreeControllerV3.PositionState.On;
            hands[1].currentPositionState = FreeControllerV3.PositionState.On;
            hands[0].currentRotationState = FreeControllerV3.RotationState.On;
            hands[1].currentRotationState = FreeControllerV3.RotationState.On;

            chest.currentPositionState = FreeControllerV3.PositionState.On;
            chest.currentRotationState = FreeControllerV3.RotationState.On;

            abdomen2.currentPositionState = FreeControllerV3.PositionState.Off;
            abdomen2.currentRotationState = FreeControllerV3.RotationState.Off;

            hip.currentPositionState = FreeControllerV3.PositionState.On;
            hip.currentRotationState = FreeControllerV3.RotationState.On;

            pelvis.currentPositionState = FreeControllerV3.PositionState.Off;
            pelvis.currentRotationState = FreeControllerV3.RotationState.Off;

            abdomen1.currentPositionState = FreeControllerV3.PositionState.Off;
            abdomen1.currentRotationState = FreeControllerV3.RotationState.Off;

            thighs[0].currentPositionState = FreeControllerV3.PositionState.Off;
            thighs[1].currentPositionState = FreeControllerV3.PositionState.Off;
            thighs[0].currentRotationState = FreeControllerV3.RotationState.Off;
            thighs[1].currentRotationState = FreeControllerV3.RotationState.Off;

            knees[0].currentPositionState = FreeControllerV3.PositionState.Off;
            knees[1].currentPositionState = FreeControllerV3.PositionState.Off;
            knees[0].currentRotationState = FreeControllerV3.RotationState.Off;
            knees[1].currentRotationState = FreeControllerV3.RotationState.Off;

            feet[0].currentPositionState = FreeControllerV3.PositionState.On;
            feet[1].currentPositionState = FreeControllerV3.PositionState.On;
            feet[0].currentRotationState = FreeControllerV3.RotationState.On;
            feet[1].currentRotationState = FreeControllerV3.RotationState.On;

            toes[0].currentPositionState = FreeControllerV3.PositionState.Off;
            toes[1].currentPositionState = FreeControllerV3.PositionState.Off;
            toes[0].currentRotationState = FreeControllerV3.RotationState.Off;
            toes[1].currentRotationState = FreeControllerV3.RotationState.Off;

            penis[0].currentPositionState = FreeControllerV3.PositionState.Off;
            penis[1].currentPositionState = FreeControllerV3.PositionState.Off;
            penis[2].currentPositionState = FreeControllerV3.PositionState.Off;
            penis[0].currentRotationState = FreeControllerV3.RotationState.Off;
            penis[1].currentRotationState = FreeControllerV3.RotationState.Off;
            penis[2].currentRotationState = FreeControllerV3.RotationState.Off;

            if (activeNodes.Count > 0)
                PHUI.UpdateStates(activeNodes.Last());
        }
        public static void MirrorNode(Node node)
        {
            PHUndo.StoreState();
            node.controller.currentPositionState = FreeControllerV3.PositionState.On;
            node.controller.currentRotationState = FreeControllerV3.RotationState.On;
            node.controller.MoveControl(GetInversePosition(node.altNode.controller.transform));
            node.controller.RotateTo(GetInverseRotation(node.altNode.controller.transform));
            node.controller.currentPositionState = node.altNode.controller.currentPositionState;
            node.controller.currentRotationState = node.altNode.controller.currentRotationState;
        }
        public static void SwapNode(Node node)
        {
            PHUndo.StoreState();
            Vector3 tempPos = GetInversePosition(node.controller.transform);
            Vector3 tempRot = GetInverseRotation(node.controller.transform).eulerAngles;
            FreeControllerV3.PositionState tempPState = node.controller.currentPositionState;
            FreeControllerV3.RotationState temRPState = node.controller.currentRotationState;
            node.controller.MoveControl(GetInversePosition(node.altNode.controller.transform));
            node.controller.RotateControl(GetInversePosition(node.altNode.controller.transform));
            node.controller.currentPositionState = node.altNode.controller.currentPositionState;
            node.controller.currentRotationState = node.altNode.controller.currentRotationState;
            node.altNode.controller.MoveControl(tempPos);
            node.altNode.controller.RotateControl(tempRot);
            node.altNode.controller.currentPositionState = tempPState;
            node.altNode.controller.currentRotationState = temRPState;
        }
        public static void PreMoveControllers()
        {
            PHUndo.StoreState();
            currentPStates = new FreeControllerV3.PositionState[activeNodes.Count];
            currentRStates = new FreeControllerV3.RotationState[activeNodes.Count];
            for (int i = 0; i < activeNodes.Count; i++)
            {
                currentPStates[i] = activeNodes[i].controller.currentPositionState;
                currentRStates[i] = activeNodes[i].controller.currentRotationState;
            }
        }
        public static void MoveActiveNodes(Vector3 dir, bool isChaingingPosition)
        {
            dir *= Time.deltaTime * PH.sensetivity.valNoCallback;
            Transform motion = Input.GetKey(KeyCode.LeftShift) ? motionTransform[1] : Input.GetKey(KeyCode.LeftControl) ? motionTransform[2] : motionTransform[motionTransformIndex];
            foreach (var node in activeNodes)
            {
                if (isChaingingPosition)
                {
                    node.controller.currentPositionState = FreeControllerV3.PositionState.On;
                    node.controller.MoveControlRelatve(motion.right * (dir.x * xPosMultiplier) + motion.up * (dir.y * yPosMultiplier) + motion.forward * (dir.z * zPosMultiplier));
                    if (PHUI.isSymmetryToggled && node.altNode != null)
                    {
                        node.altNode.controller.currentPositionState = FreeControllerV3.PositionState.On;
                        node.altNode.controller.MoveControlRelatve(motion.right * (dir.x * -xPosMultiplier) + motion.up * (dir.y * yPosMultiplier) + motion.forward * (dir.z * zPosMultiplier));
                    }
                }
                else
                {
                    node.controller.currentRotationState = FreeControllerV3.RotationState.On;
                    node.controller.RotateControl(motion.right * (dir.x * xRotMultiplier) + motion.up * (dir.y * yRotMultiplier) + motion.forward * (dir.z * zRotMultiplier), PH.rotationSpeed.val);
                    if (PHUI.isSymmetryToggled && node.altNode != null)
                    {
                        node.altNode.controller.currentRotationState = FreeControllerV3.RotationState.On;
                        node.altNode.controller.RotateControl(motion.right * (dir.x * xRotMultiplier) + motion.up * (dir.y * -yRotMultiplier) + motion.forward * (dir.z * -zRotMultiplier), PH.rotationSpeed.val);
                    }
                }
            }

        }
        public static void PostMoveControllers()
        {
            for (int i = 0; i < activeNodes.Count; i++)
            {
                activeNodes[i].controller.currentPositionState = currentPStates[i];
                activeNodes[i].controller.currentRotationState = currentRStates[i];
                if (PHUI.isSymmetryToggled && activeNodes[i].altNode != null)
                {
                    activeNodes[i].altNode.controller.currentPositionState = currentPStates[i];
                    activeNodes[i].altNode.controller.currentRotationState = currentRStates[i];
                }
            }
        }

        //public static void MirrorArmsMethod(bool right)
        //{
        //    int mirror = right ? 0 : 1; int mirrored = right ? 1 : 0;
        //    //shoulders[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
        //    //shoulders[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
        //    //elbows[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
        //    //elbows[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
        //    //hands[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
        //    //hands[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
        //    shoulders[mirrored].MoveControl(GetInversePosition(shoulders[mirror].transform));
        //    elbows[mirrored].MoveControl(GetInversePosition(elbows[mirror].transform));
        //    hands[mirrored].MoveControl(GetInversePosition(hands[mirror].transform));
        //    shoulders[mirrored].RotateTo(GetInverseRotation(shoulders[mirror].transform));
        //    elbows[mirrored].RotateTo(GetInverseRotation(elbows[mirror].transform));
        //    hands[mirrored].RotateTo(GetInverseRotation(hands[mirror].transform));
        //    shoulders[mirrored].currentPositionState = shoulders[mirror].currentPositionState;
        //    elbows[mirrored].currentPositionState = elbows[mirror].currentPositionState;
        //    hands[mirrored].currentPositionState = hands[mirror].currentPositionState;
        //    shoulders[mirrored].currentRotationState = shoulders[mirror].currentRotationState;
        //    elbows[mirrored].currentRotationState = elbows[mirror].currentRotationState;
        //    hands[mirrored].currentRotationState = hands[mirror].currentRotationState;
        //}

        //public static void MirrorLegsMethod(bool right)
        //{
        //    int mirror = right ? 0 : 1; int mirrored = right ? 1 : 0;
        //    //thighs[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
        //    //thighs[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
        //    //knees[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
        //    //knees[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
        //    //feet[mirrored].currentPositionState = FreeControllerV3.PositionState.On;
        //    //feet[mirrored].currentRotationState = FreeControllerV3.RotationState.On;
        //    thighs[mirrored].MoveControl(GetInversePosition(thighs[mirror].transform));
        //    knees[mirrored].MoveControl(GetInversePosition(knees[mirror].transform));
        //    feet[mirrored].MoveControl(GetInversePosition(feet[mirror].transform));
        //    toes[mirrored].MoveControl(GetInversePosition(toes[mirror].transform));
        //    thighs[mirrored].RotateTo(GetInverseRotation(thighs[mirror].transform));
        //    knees[mirrored].RotateTo(GetInverseRotation(knees[mirror].transform));
        //    feet[mirrored].RotateTo(GetInverseRotation(feet[mirror].transform));
        //    toes[mirrored].RotateTo(GetInverseRotation(toes[mirror].transform));
        //    thighs[mirrored].currentPositionState = thighs[mirror].currentPositionState;
        //    knees[mirrored].currentPositionState = knees[mirror].currentPositionState;
        //    feet[mirrored].currentPositionState = feet[mirror].currentPositionState;
        //    toes[mirrored].currentPositionState = toes[mirror].currentPositionState;
        //    thighs[mirrored].currentRotationState = thighs[mirror].currentRotationState;
        //    knees[mirrored].currentRotationState = knees[mirror].currentRotationState;
        //    feet[mirrored].currentRotationState = feet[mirror].currentRotationState;
        //    toes[mirrored].currentRotationState = toes[mirror].currentRotationState;
        //}
        static Vector3 GetInversePosition(Transform trans)
        {
            Vector3 inversePos = Vector3.Reflect(trans.position - activeAtom.freeControllers[0].transform.position, activeAtom.freeControllers[0].transform.right);
            return activeAtom.freeControllers[0].transform.position + inversePos;
        }
        static Quaternion GetInverseRotation(Transform trans)
        {
            return ReflectRotation(trans.rotation, activeAtom.freeControllers[0].transform.right);
        }
        static Quaternion ReflectRotation(Quaternion source, Vector3 normal)
        {
            return Quaternion.LookRotation(Vector3.Reflect(source * Vector3.forward, normal), Vector3.Reflect(source * Vector3.up, normal));
        }
        public static void ClearLink()
        {
            foreach (var node in activeNodes)
            {
                node.controller.SelectLinkToRigidbody(null);
                node.controller.currentPositionState = FreeControllerV3.PositionState.On;
                node.controller.currentRotationState = FreeControllerV3.RotationState.On;
            }
            PHUI.UpdateStates(activeNodes.Last());
        }
        public static void IKLink()
        {
            switch (activeNodes.Count)
            {
                case 0:
                    IKLink(pelvis);
                    break;
                case 1:
                    if (activeNodes[0].controller.name == activeAtom.freeControllers[0].name)
                        IKLink(pelvis);
                    else
                        IKLink(activeNodes[0].controller);
                    break;
                case 2:
                    if (Input.GetKey(KeyCode.LeftAlt))
                        Link(activeNodes[0].controller, activeNodes[1].controller);
                    else
                        IKLink(activeNodes[0].controller, activeNodes[1].controller);
                    break;
                case 3:
                    Link(activeNodes[0].controller, activeNodes[2].controller);
                    Link(activeNodes[1].controller, activeNodes[2].controller);
                    break;
                default:
                    return;
            }
        }
        static void IKLink(FreeControllerV3 a, FreeControllerV3 b)
        {
            PHUndo.StoreState();
            PHLinker.LinkPath(PHLinker.FindPath(a, b));
        }
        static void IKLink(FreeControllerV3 center)
        {
            PHLinker.LinkPath(PHLinker.FindPath(toes[0], center));
            PHLinker.LinkPath(PHLinker.FindPath(toes[1], center));
            PHLinker.LinkPath(PHLinker.FindPath(head, center));
            PHLinker.LinkPath(PHLinker.FindPath(penis[2], center));
            PHLinker.LinkPath(PHLinker.FindPath(hands[1], center));
            PHLinker.LinkPath(PHLinker.FindPath(hands[0], center));
            SetDefaultLinkState(center);
        }
        static void SetDefaultLinkState(FreeControllerV3 center)
        {
            if (Input.GetKey(KeyCode.LeftShift))
            {
                head.currentPositionState = FreeControllerV3.PositionState.Off;
                penis[1].currentPositionState = FreeControllerV3.PositionState.Off;
                penis[2].currentPositionState = FreeControllerV3.PositionState.Off;
                for (int i = 0; i < 2; i++)
                {
                    shoulders[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    elbows[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    hands[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    thighs[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    knees[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    feet[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    toes[i].currentPositionState = FreeControllerV3.PositionState.Off;
                }
            }
            center.currentPositionState = FreeControllerV3.PositionState.On;
            center.currentRotationState = FreeControllerV3.RotationState.On;
        }
        public static void Link(FreeControllerV3 a, FreeControllerV3 b)
        {
            a.SelectLinkToRigidbody(b.gameObject.GetComponent<Rigidbody>());
            a.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            a.currentRotationState = FreeControllerV3.RotationState.ParentLink;
        }
        static void HandleLoadPreset(string path)
        {
            var poseJson = JSON.Parse(SuperController.singleton.ReadFileIntoString(path));
            if (!bool.Parse(poseJson["V2"].Value))
            {
                SuperController.LogError("older pose files are not supported");
                return;
            }
            foreach (var fc in activeAtom.freeControllers)
            {
                fc.Reset();
                fc.collisionEnabled = false;
            }
            var JSONControllers = poseJson["storables"].AsArray;
            var morphControl = ((activeAtom.GetStorableByID("geometry")) as DAZCharacterSelector).morphsControlUI;
            foreach (var JSONController in JSONControllers.Childs)
            {
                if (!JSONController["id"].Value.Contains("Control"))
                    continue;

                var controller = activeAtom.freeControllers.First(fc => fc.name == JSONController["id"].Value);

                if (controller == null)
                    continue;

                controller.physicsEnabled = false;
                SetBoolParam(controller, JSONController, "detachControl");
                SetBoolParam(controller, JSONController, "interactableInPlayMode");
                SetBoolParam(controller, JSONController, "deactivateOtherControlsOnPossess");
                SetBoolParam(controller, JSONController, "possessable");
                SetBoolParam(controller, JSONController, "canGrabPosition");
                SetBoolParam(controller, JSONController, "canGrabRotation");
                SetBoolParam(controller, JSONController, "freezeAtomPhysicsWhenGrabbed");
                SetBoolParam(controller, JSONController, "xPositionLock");
                SetBoolParam(controller, JSONController, "yPositionLock");
                SetBoolParam(controller, JSONController, "zPositionLock");
                SetBoolParam(controller, JSONController, "xPositionLocalLock");
                SetBoolParam(controller, JSONController, "yPositionLocalLock");
                SetBoolParam(controller, JSONController, "zPositionLocalLock");
                SetBoolParam(controller, JSONController, "xRotationLock");
                SetBoolParam(controller, JSONController, "yRotationLock");
                SetBoolParam(controller, JSONController, "zRotationLock");
                SetBoolParam(controller, JSONController, "physicsEnabled");
                SetBoolParam(controller, JSONController, "useGravity");
                SetBoolParam(controller, JSONController, "maxVelocityEnable");

                SetFloatParam(controller, JSONController, "complyPositionThreshold");
                SetFloatParam(controller, JSONController, "complyRotationThreshold");
                SetFloatParam(controller, JSONController, "complySpeed");
                SetFloatParam(controller, JSONController, "positionGrid");
                SetFloatParam(controller, JSONController, "rotationGrid");
                SetFloatParam(controller, JSONController, "mass");
                SetFloatParam(controller, JSONController, "drag");
                SetFloatParam(controller, JSONController, "maxVelocity");
                SetFloatParam(controller, JSONController, "angularDrag");
                SetFloatParam(controller, JSONController, "holdPositionSpring");
                SetFloatParam(controller, JSONController, "holdPositionDamper");
                SetFloatParam(controller, JSONController, "holdPositionMaxForce");
                SetFloatParam(controller, JSONController, "holdRotationSpring");
                SetFloatParam(controller, JSONController, "holdRotationDamper");
                SetFloatParam(controller, JSONController, "holdRotationMaxForce");
                SetFloatParam(controller, JSONController, "complyPositionSpring");
                SetFloatParam(controller, JSONController, "complyPositionDamper");
                SetFloatParam(controller, JSONController, "complyRotationSpring");
                SetFloatParam(controller, JSONController, "complyRotationDamper");
                SetFloatParam(controller, JSONController, "linkPositionSpring");
                SetFloatParam(controller, JSONController, "linkPositionDamper");
                SetFloatParam(controller, JSONController, "linkPositionMaxForce");
                SetFloatParam(controller, JSONController, "linkRotationSpring");
                SetFloatParam(controller, JSONController, "linkRotationDamper");
                SetFloatParam(controller, JSONController, "linkRotationMaxForce");
                SetFloatParam(controller, JSONController, "complyJointDriveSpring");
                SetFloatParam(controller, JSONController, "jointDriveSpring");
                SetFloatParam(controller, JSONController, "jointDriveDamper");
                SetFloatParam(controller, JSONController, "jointDriveMaxForce");
                SetFloatParam(controller, JSONController, "jointDriveXTarget");
                SetFloatParam(controller, JSONController, "jointDriveYTarget");
                SetFloatParam(controller, JSONController, "jointDriveZTarget");

                SetStringParam(controller, JSONController, "positionState");
                SetStringParam(controller, JSONController, "rotationState");
                SetStringParam(controller, JSONController, "positionGridMode");
                SetStringParam(controller, JSONController, "rotationGridMode");

                controller.transform.localPosition = SetVec3Param(JSONController, "localPosition");

                controller.transform.localEulerAngles = SetVec3Param(JSONController, "localRotation");

                controller.physicsEnabled = true;
            }
            foreach (var fc in activeAtom.freeControllers)
            {
                fc.collisionEnabled = true;
            }
            if (PH.changePose.valNoCallback)
            {
                morphControl.GetMorphUids().ForEach((uid) =>
                {
                    var morph = morphControl.GetMorphByUid(uid);
                    if (morph.isPoseControl)
                    {
                        morph.SetDefaultValue();
                    }
                });
                var geoNode = JSONControllers.Childs.FirstOrDefault(node => node["id"].Value == "geometry");
                if (geoNode == null)
                    return;
                foreach (var morph in geoNode["morphs"].Childs)
                {
                    morphControl.GetMorphByDisplayName(morph["name"].Value).morphValue = float.Parse(morph["value"].Value);
                }
            }

        }
        static void SetBoolParam(FreeControllerV3 controller, JSONNode node, string name)
        {
            if (controller.GetBoolParamNames().Contains(name))
                controller.SetBoolParamValue(name, bool.Parse(node[name].Value));
        }
        static void SetFloatParam(FreeControllerV3 controller, JSONNode node, string name)
        {
            if (controller.GetFloatParamNames().Contains(name))
                controller.SetFloatParamValue(name, float.Parse(node[name].Value));
        }
        static void SetStringParam(FreeControllerV3 controller, JSONNode node, string name)
        {
            if (controller.GetStringParamNames().Contains(name))
                controller.SetStringParamValue(name, node[name].Value);
        }
        static Vector3 SetVec3Param(JSONNode node, string name)
        {
            float x, y, z;
            x = float.Parse((node[name]["x"].Value));
            y = float.Parse((node[name]["y"].Value));
            z = float.Parse((node[name]["z"].Value));
            return new Vector3(x, y, z);
        }
    }
}
